# encoding: utf-8

"""Test suite for pptx.dml.fill module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from pptx.dml.color import ColorFormat
from pptx.dml.fill import (
    _BlipFill,
    _Fill,
    FillFormat,
    _GradFill,
    _GradientStop,
    _GradientStops,
    _GrpFill,
    _NoFill,
    _NoneFill,
    _PattFill,
    _SolidFill,
)
from pptx.enum.dml import MSO_FILL, MSO_PATTERN
from pptx.oxml.dml.fill import CT_GradientStopList

from ..unitutil.cxml import element, xml
from ..unitutil.mock import class_mock, instance_mock, method_mock, property_mock


class DescribeFillFormat(object):
    def it_can_set_the_fill_type_to_no_fill(self, background_fixture):
        fill, _NoFill_, expected_xml, no_fill_ = background_fixture

        fill.background()

        assert fill._xPr.xml == expected_xml
        _NoFill_.assert_called_once_with(fill._xPr.eg_fillProperties)
        assert fill._fill is no_fill_

    def it_can_set_the_fill_type_to_gradient(self, gradient_fixture):
        fill, _GradFill_, expected_xml, grad_fill_ = gradient_fixture

        fill.gradient()

        assert fill._xPr.xml == expected_xml
        _GradFill_.assert_called_once_with(fill._xPr.eg_fillProperties)
        assert fill._fill is grad_fill_

    def it_can_set_the_fill_type_to_solid(self, solid_fixture):
        fill, _SolidFill_, expected_xml, solid_fill_ = solid_fixture

        fill.solid()

        assert fill._xPr.xml == expected_xml
        _SolidFill_.assert_called_once_with(fill._xPr.eg_fillProperties)
        assert fill._fill is solid_fill_

    def it_can_set_the_fill_type_to_patterned(self, patterned_fixture):
        fill, _PattFill_, expected_xml, patt_fill_ = patterned_fixture

        fill.patterned()

        assert fill._xPr.xml == expected_xml
        _PattFill_.assert_called_once_with(fill._xPr.eg_fillProperties)
        assert fill._fill is patt_fill_

    def it_provides_access_to_its_foreground_color(self, fore_color_fixture):
        fill, color_ = fore_color_fixture
        color = fill.fore_color
        assert color is color_

    def it_provides_access_to_its_background_color(self, back_color_fixture):
        fill, color_ = back_color_fixture
        color = fill.back_color
        assert color is color_

    def it_knows_its_fill_type(self, fill_type_fixture):
        fill, expected_value = fill_type_fixture
        fill_type = fill.type
        assert fill_type == expected_value

    def it_knows_the_angle_of_a_linear_gradient(self, grad_fill_, type_prop_):
        grad_fill_.gradient_angle = 42.0
        type_prop_.return_value = MSO_FILL.GRADIENT
        fill = FillFormat(None, grad_fill_)

        angle = fill.gradient_angle

        assert angle == 42.0

    def it_can_change_the_angle_of_a_linear_gradient(self, grad_fill_, type_prop_):
        type_prop_.return_value = MSO_FILL.GRADIENT
        fill = FillFormat(None, grad_fill_)

        fill.gradient_angle = 42.24

        assert grad_fill_.gradient_angle == 42.24

    def it_provides_access_to_the_gradient_stops(
        self, type_prop_, grad_fill_, gradient_stops_
    ):
        type_prop_.return_value = MSO_FILL.GRADIENT
        grad_fill_.gradient_stops = gradient_stops_
        fill = FillFormat(None, grad_fill_)

        gradient_stops = fill.gradient_stops

        assert gradient_stops is gradient_stops_

    def it_raises_on_non_gradient_fill(self, grad_fill_, type_prop_):
        type_prop_.return_value = None
        fill = FillFormat(None, grad_fill_)
        with pytest.raises(TypeError):
            fill.gradient_angle
        with pytest.raises(TypeError):
            fill.gradient_angle = 123.4
        with pytest.raises(TypeError):
            fill.gradient_stops

    def it_knows_its_pattern(self, pattern_get_fixture):
        fill, expected_value = pattern_get_fixture
        pattern = fill.pattern
        assert pattern == expected_value

    def it_can_change_its_pattern(self, pattern_set_fixture):
        fill, pattern = pattern_set_fixture
        fill.pattern = pattern
        assert fill.pattern is pattern

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def back_color_fixture(self, fill_, color_):
        fill_.back_color = color_
        fill_format = FillFormat(None, fill_)
        return fill_format, color_

    @pytest.fixture(
        params=[
            ("p:spPr{a:b=c}", "p:spPr{a:b=c}/a:noFill"),
            ("a:tcPr/a:blipFill", "a:tcPr/a:noFill"),
            ("a:rPr/a:gradFill", "a:rPr/a:noFill"),
            ("a:tcPr/a:grpFill", "a:tcPr/a:noFill"),
            ("a:defRPr/a:noFill", "a:defRPr/a:noFill"),
            ("a:endParaRPr/a:pattFill", "a:endParaRPr/a:noFill"),
            ("a:rPr/a:solidFill", "a:rPr/a:noFill"),
        ]
    )
    def background_fixture(self, request, no_fill_):
        cxml, expected_cxml = request.param

        fill = FillFormat.from_fill_parent(element(cxml))
        # --mock must be after FillFormat call to avoid poss. contructor call
        _NoFill_ = class_mock(
            request, "pptx.dml.fill._NoFill", return_value=no_fill_, autospec=True
        )
        expected_xml = xml(expected_cxml)
        return fill, _NoFill_, expected_xml, no_fill_

    @pytest.fixture
    def fill_type_fixture(self, fill_):
        expected_value = fill_.type = 42
        fill = FillFormat(None, fill_)
        return fill, expected_value

    @pytest.fixture
    def fore_color_fixture(self, fill_, color_):
        fill_.fore_color = color_
        fill_format = FillFormat(None, fill_)
        return fill_format, color_

    @pytest.fixture(
        params=[
            ("p:spPr{a:b=c}", "p:spPr{a:b=c}/a:gradFill"),
            ("p:bgPr/a:noFill", "p:bgPr/a:gradFill"),
            ("a:tcPr/a:blipFill", "a:tcPr/a:gradFill"),
            ("a:rPr/a:grpFill", "a:rPr/a:gradFill"),
            ("a:tcPr/a:noFill", "a:tcPr/a:gradFill"),
            ("a:defRPr/a:solidFill", "a:defRPr/a:gradFill"),
            ("a:endParaRPr/a:blipFill", "a:endParaRPr/a:gradFill"),
        ]
    )
    def gradient_fixture(self, request, grad_fill_):
        cxml, expected_cxml = request.param

        fill = FillFormat.from_fill_parent(element(cxml))
        # --mock must be after FillFormat call to avoid poss. contructor call
        _GradFill_ = class_mock(
            request, "pptx.dml.fill._GradFill", return_value=grad_fill_, autospec=True
        )
        expected_xml = xml(
            expected_cxml + "{rotWithShape=1}/(a:gsLst/(a:gs{pos=0}/a:scheme"
            "Clr{val=accent1}/(a:tint{val=100000},a:shade{val=100000},a:satM"
            "od{val=130000}),a:gs{pos=100000}/a:schemeClr{val=accent1}/(a:ti"
            "nt{val=50000},a:shade{val=100000},a:satMod{val=350000})),a:lin{"
            "scaled=0})"
        )
        return fill, _GradFill_, expected_xml, grad_fill_

    @pytest.fixture
    def pattern_get_fixture(self, fill_):
        expected_value = fill_.pattern = MSO_PATTERN.WAVE
        fill = FillFormat(None, fill_)
        return fill, expected_value

    @pytest.fixture
    def pattern_set_fixture(self, fill_):
        pattern = MSO_PATTERN.DIVOT
        fill = FillFormat(None, fill_)
        return fill, pattern

    @pytest.fixture(
        params=[
            ("p:spPr{a:b=c}", "p:spPr{a:b=c}/a:pattFill"),
            ("a:tcPr/a:blipFill", "a:tcPr/a:pattFill"),
            ("a:rPr/a:gradFill", "a:rPr/a:pattFill"),
            ("a:tcPr/a:grpFill", "a:tcPr/a:pattFill"),
            ("a:defRPr/a:noFill", "a:defRPr/a:pattFill"),
            ("a:endParaRPr/a:pattFill", "a:endParaRPr/a:pattFill"),
            ("a:rPr/a:solidFill", "a:rPr/a:pattFill"),
        ]
    )
    def patterned_fixture(self, request, patt_fill_):
        cxml, expected_cxml = request.param

        fill = FillFormat.from_fill_parent(element(cxml))
        # --mock must be after FillFormat call to avoid poss. contructor call
        _PattFill_ = class_mock(
            request, "pptx.dml.fill._PattFill", return_value=patt_fill_, autospec=True
        )
        expected_xml = xml(expected_cxml)
        return fill, _PattFill_, expected_xml, patt_fill_

    @pytest.fixture(
        params=[
            ("p:spPr{a:b=c}", "p:spPr{a:b=c}/a:solidFill"),
            ("a:tcPr/a:blipFill", "a:tcPr/a:solidFill"),
            ("a:rPr/a:gradFill", "a:rPr/a:solidFill"),
            ("a:tcPr/a:grpFill", "a:tcPr/a:solidFill"),
            ("a:defRPr/a:noFill", "a:defRPr/a:solidFill"),
            ("a:rPr/a:pattFill", "a:rPr/a:solidFill"),
            ("a:endParaRPr/a:solidFill", "a:endParaRPr/a:solidFill"),
        ]
    )
    def solid_fixture(self, request, solid_fill_):
        cxml, expected_cxml = request.param

        fill = FillFormat.from_fill_parent(element(cxml))
        # --mock must be after FillFormat call to avoid poss. contructor call
        _SolidFill_ = class_mock(
            request, "pptx.dml.fill._SolidFill", return_value=solid_fill_, autospec=True
        )
        expected_xml = xml(expected_cxml)
        return fill, _SolidFill_, expected_xml, solid_fill_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def color_(self, request):
        return instance_mock(request, ColorFormat)

    @pytest.fixture
    def fill_(self, request):
        return instance_mock(request, _Fill)

    @pytest.fixture
    def grad_fill_(self, request):
        return instance_mock(request, _GradFill)

    @pytest.fixture
    def gradient_stops_(self, request):
        return instance_mock(request, _GradientStops)

    @pytest.fixture
    def no_fill_(self, request):
        return instance_mock(request, _NoFill)

    @pytest.fixture
    def patt_fill_(self, request):
        return instance_mock(request, _PattFill)

    @pytest.fixture
    def solid_fill_(self, request):
        return instance_mock(request, _SolidFill)

    @pytest.fixture
    def type_prop_(self, request):
        return property_mock(request, FillFormat, "type")


class Describe_Fill(object):
    def it_raises_on_fore_color_access(self, fore_raise_fixture):
        fill, exception_type = fore_raise_fixture
        with pytest.raises(exception_type):
            fill.fore_color

    def it_raises_on_back_color_access(self, back_raise_fixture):
        fill, exception_type = back_raise_fixture
        with pytest.raises(exception_type):
            fill.back_color

    def it_raises_on_pattern_access(self, pattern_raise_fixture):
        fill, exception_type = pattern_raise_fixture
        with pytest.raises(exception_type):
            fill.pattern

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def back_raise_fixture(self):
        fill = _Fill("foobar")
        exception_type = TypeError
        return fill, exception_type

    @pytest.fixture
    def fore_raise_fixture(self):
        fill = _Fill("foobar")
        exception_type = TypeError
        return fill, exception_type

    @pytest.fixture
    def pattern_raise_fixture(self):
        fill = _Fill("barfoo")
        exception_type = TypeError
        return fill, exception_type


class Describe_BlipFill(object):
    def it_knows_its_fill_type(self, fill_type_fixture):
        blip_fill, expected_value = fill_type_fixture
        fill_type = blip_fill.type
        assert fill_type == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def fill_type_fixture(self):
        xFill = element("a:blipFill")
        blip_fill = _BlipFill(xFill)
        expected_value = MSO_FILL.PICTURE
        return blip_fill, expected_value


class Describe_GradFill(object):
    def it_knows_the_gradient_angle(self, angle_fixture):
        grad_fill, expected_value = angle_fixture
        angle = grad_fill.gradient_angle
        assert angle == expected_value

    def it_can_change_the_gradient_angle(self, angle_set_fixture):
        grad_fill, angle, expected_xml = angle_set_fixture
        grad_fill.gradient_angle = angle
        assert grad_fill._gradFill.xml == expected_xml

    def it_provides_access_to_the_gradient_stops(self, stops_fixture):
        grad_fill, _GradientStops_ = stops_fixture[:2]
        expected_xml, gradient_stops_ = stops_fixture[2:]

        gradient_stops = grad_fill.gradient_stops

        gradFill = grad_fill._gradFill
        _GradientStops_.assert_called_once_with(gradFill[0])
        assert gradFill.xml == expected_xml
        assert gradient_stops is gradient_stops_

    def it_raises_on_non_linear_gradient(self):
        gradFill = element("a:gradFill/a:path")
        grad_fill = _GradFill(gradFill)
        with pytest.raises(ValueError):
            grad_fill.gradient_angle
        with pytest.raises(ValueError):
            grad_fill.gradient_angle = 43.21

    def it_knows_its_fill_type(self, fill_type_fixture):
        grad_fill, expected_value = fill_type_fixture
        fill_type = grad_fill.type
        assert fill_type == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("a:gradFill", None),
            ("a:gradFill/a:lin{ang=0}", 0.0),
            ("a:gradFill/a:lin{ang=2730000}", 314.5),
            ("a:gradFill/a:lin{ang=16200000}", 90.0),
        ]
    )
    def angle_fixture(self, request):
        cxml, expected_value = request.param
        grad_fill = _GradFill(element(cxml))
        return grad_fill, expected_value

    @pytest.fixture(
        params=[
            (301.2, "a:gradFill/a:lin{ang=3528000}"),
            (31.22, "a:gradFill/a:lin{ang=19726800}"),
            (0, "a:gradFill/a:lin{ang=0}"),
            (-460.0, "a:gradFill/a:lin{ang=6000000}"),
        ]
    )
    def angle_set_fixture(self, request):
        angle, expected_cxml = request.param
        grad_fill = _GradFill(element("a:gradFill/a:lin{ang=0}"))
        expected_xml = xml(expected_cxml)
        return grad_fill, angle, expected_xml

    @pytest.fixture
    def fill_type_fixture(self):
        xFill = element("a:gradFill")
        grad_fill = _GradFill(xFill)
        expected_value = MSO_FILL.GRADIENT
        return grad_fill, expected_value

    @pytest.fixture(
        params=[
            ("a:gradFill", "a:gradFill", True),
            ("a:gradFill/a:gsLst", "a:gradFill/a:gsLst", False),
        ]
    )
    def stops_fixture(self, request, _GradientStops_, gradient_stops_):
        gradFill_cxml, expected_cxml, add_gsLst = request.param
        _GradientStops_.return_value = gradient_stops_
        grad_fill = _GradFill(element(gradFill_cxml))
        if add_gsLst:
            expected_xml = xml(
                expected_cxml + "/a:gsLst/(a:gs{pos=0}/a:schemeClr{val=accen"
                "t1}/(a:tint{val=100000},a:shade{val=100000},a:satMod{val=13"
                "0000}),a:gs{pos=100000}/a:schemeClr{val=accent1}/(a:tint{va"
                "l=50000},a:shade{val=100000},a:satMod{val=350000}))"
            )
        else:
            expected_xml = xml(expected_cxml)
        return grad_fill, _GradientStops_, expected_xml, gradient_stops_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _GradientStops_(self, request):
        return class_mock(request, "pptx.dml.fill._GradientStops")

    @pytest.fixture
    def gradient_stops_(self, request):
        return instance_mock(request, _GradientStops)


class Describe_GrpFill(object):
    def it_knows_its_fill_type(self, fill_type_fixture):
        grp_fill, expected_value = fill_type_fixture
        fill_type = grp_fill.type
        assert fill_type == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def fill_type_fixture(self):
        xFill = element("a:grpFill")
        grp_fill = _GrpFill(xFill)
        expected_value = MSO_FILL.GROUP
        return grp_fill, expected_value


class Describe_NoFill(object):
    def it_knows_its_fill_type(self, fill_type_fixture):
        no_fill, expected_value = fill_type_fixture
        fill_type = no_fill.type
        assert fill_type == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def fill_type_fixture(self):
        xFill = element("a:noFill")
        no_fill = _NoFill(xFill)
        expected_value = MSO_FILL.BACKGROUND
        return no_fill, expected_value


class Describe_NoneFill(object):
    def it_knows_its_fill_type(self, fill_type_fixture):
        none_fill, expected_value = fill_type_fixture
        fill_type = none_fill.type
        assert fill_type == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def fill_type_fixture(self):
        none_fill = _NoneFill(None)
        expected_value = None
        return none_fill, expected_value


class Describe_PattFill(object):
    def it_knows_its_fill_type(self, fill_type_fixture):
        patt_fill, expected_value = fill_type_fixture
        fill_type = patt_fill.type
        assert fill_type == expected_value

    def it_knows_its_pattern(self, pattern_get_fixture):
        patt_fill, expected_value = pattern_get_fixture
        pattern = patt_fill.pattern
        assert pattern == expected_value

    def it_can_change_its_pattern(self, pattern_set_fixture):
        patt_fill, pattern, pattFill, expected_xml = pattern_set_fixture
        patt_fill.pattern = pattern
        assert pattFill.xml == expected_xml

    def it_provides_access_to_its_foreground_color(self, fore_color_fixture):
        patt_fill, pattFill, expected_xml, color_ = fore_color_fixture[:4]
        ColorFormat_from_colorchoice_parent_ = fore_color_fixture[4]

        color = patt_fill.fore_color

        assert pattFill.xml == expected_xml
        ColorFormat_from_colorchoice_parent_.assert_called_once_with(pattFill.fgClr)
        assert color is color_

    def it_provides_access_to_its_background_color(self, back_color_fixture):
        patt_fill, pattFill, expected_xml, color_ = back_color_fixture[:4]
        ColorFormat_from_colorchoice_parent_ = back_color_fixture[4]

        color = patt_fill.back_color

        assert pattFill.xml == expected_xml
        ColorFormat_from_colorchoice_parent_.assert_called_once_with(pattFill.bgClr)
        assert color is color_

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("a:pattFill", "a:pattFill/a:bgClr/a:srgbClr{val=FFFFFF}"),
            ("a:pattFill/a:bgClr", "a:pattFill/a:bgClr"),
            (
                "a:pattFill/a:bgClr/a:schemeClr{val=accent1}",
                "a:pattFill/a:bgClr/a:schemeClr{val=accent1}",
            ),
        ]
    )
    def back_color_fixture(self, request, color_, ColorFormat_from_colorchoice_parent_):
        pattFill_cxml, expected_cxml = request.param
        pattFill = element(pattFill_cxml)
        expected_xml = xml(expected_cxml)
        ColorFormat_from_colorchoice_parent_.return_value = color_

        patt_fill = _PattFill(pattFill)
        return (
            patt_fill,
            pattFill,
            expected_xml,
            color_,
            ColorFormat_from_colorchoice_parent_,
        )

    @pytest.fixture(
        params=[
            ("a:pattFill", "a:pattFill/a:fgClr/a:srgbClr{val=000000}"),
            ("a:pattFill/a:fgClr", "a:pattFill/a:fgClr"),
            (
                "a:pattFill/a:fgClr/a:schemeClr{val=accent2}",
                "a:pattFill/a:fgClr/a:schemeClr{val=accent2}",
            ),
        ]
    )
    def fore_color_fixture(self, request, color_, ColorFormat_from_colorchoice_parent_):
        pattFill_cxml, expected_cxml = request.param
        pattFill = element(pattFill_cxml)
        expected_xml = xml(expected_cxml)
        ColorFormat_from_colorchoice_parent_.return_value = color_

        patt_fill = _PattFill(pattFill)
        return (
            patt_fill,
            pattFill,
            expected_xml,
            color_,
            ColorFormat_from_colorchoice_parent_,
        )

    @pytest.fixture
    def fill_type_fixture(self):
        patt_fill = _PattFill(element("a:pattFill"))
        expected_value = MSO_FILL.PATTERNED
        return patt_fill, expected_value

    @pytest.fixture(
        params=[
            ("a:pattFill", None),
            ("a:pattFill{prst=diagCross}", MSO_PATTERN.DIAGONAL_CROSS),
            ("a:pattFill{prst=wave}", MSO_PATTERN.WAVE),
        ]
    )
    def pattern_get_fixture(self, request):
        pattFill_cxml, expected_value = request.param
        pattFill = element(pattFill_cxml)

        patt_fill = _PattFill(pattFill)
        return patt_fill, expected_value

    @pytest.fixture(
        params=[
            ("a:pattFill", MSO_PATTERN.WAVE, "a:pattFill{prst=wave}"),
            ("a:pattFill{prst=wave}", MSO_PATTERN.DIVOT, "a:pattFill{prst=divot}"),
            ("a:pattFill{prst=divot}", None, "a:pattFill"),
        ]
    )
    def pattern_set_fixture(self, request):
        pattFill_cxml, pattern, expected_cxml = request.param
        pattFill = element(pattFill_cxml)
        expected_xml = xml(expected_cxml)

        patt_fill = _PattFill(pattFill)
        return patt_fill, pattern, pattFill, expected_xml

    # fixture components ---------------------------------------------

    @pytest.fixture
    def ColorFormat_from_colorchoice_parent_(self, request):
        return method_mock(request, ColorFormat, "from_colorchoice_parent")

    @pytest.fixture
    def color_(self, request):
        return instance_mock(request, ColorFormat)


class Describe_SolidFill(object):
    def it_knows_its_fill_type(self, fill_type_fixture):
        solid_fill, expected_value = fill_type_fixture
        fill_type = solid_fill.type
        assert fill_type == expected_value

    def it_provides_access_to_its_foreground_color(self, fore_color_fixture):
        solid_fill, solidFill, color_ = fore_color_fixture[:3]
        ColorFormat_from_colorchoice_parent_ = fore_color_fixture[3]

        color = solid_fill.fore_color

        ColorFormat_from_colorchoice_parent_.assert_called_once_with(solidFill)
        assert color is color_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def fill_type_fixture(self):
        solid_fill = _SolidFill(element("a:solidFill"))
        expected_value = MSO_FILL.SOLID
        return solid_fill, expected_value

    @pytest.fixture
    def fore_color_fixture(self, ColorFormat_from_colorchoice_parent_, color_):
        ColorFormat_from_colorchoice_parent_.return_value = color_
        solidFill = element("a:solidFill")

        solid_fill = _SolidFill(solidFill)
        return (solid_fill, solidFill, color_, ColorFormat_from_colorchoice_parent_)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def ColorFormat_from_colorchoice_parent_(self, request):
        return method_mock(request, ColorFormat, "from_colorchoice_parent")

    @pytest.fixture
    def color_(self, request):
        return instance_mock(request, ColorFormat)


class Describe_GradientStops(object):
    def it_provides_access_to_its_stops(self):
        stops = _GradientStops(CT_GradientStopList.new_gsLst())

        assert len(stops) == 2
        for stop in stops:
            assert isinstance(stop, _GradientStop)


class Describe_GradientStop(object):
    def it_provides_access_to_its_color(self, request):
        gs = element("a:gs")
        ColorFormat_ = class_mock(request, "pptx.dml.fill.ColorFormat")
        color_ = instance_mock(request, ColorFormat)
        ColorFormat_.from_colorchoice_parent.return_value = color_
        stop = _GradientStop(gs)

        color = stop.color

        ColorFormat_.from_colorchoice_parent.assert_called_once_with(gs)
        assert color is color_

    def it_knows_its_position(self, pos_get_fixture):
        stop, expected_value = pos_get_fixture
        position = stop.position
        assert position == expected_value

    def it_can_change_its_position(self, pos_set_fixture):
        stop, new_value, expected_value = pos_set_fixture
        stop.position = new_value
        assert stop.position == expected_value

    def it_raises_on_position_out_of_range(self, raises_fixture):
        stop, out_of_range_value = raises_fixture
        with pytest.raises(ValueError):
            stop.position = out_of_range_value

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("a:gs{pos=0}", 0.0),
            ("a:gs{pos=42240}", 0.4224),
            ("a:gs{pos=100000}", 1.0),
        ]
    )
    def pos_get_fixture(self, request):
        gs_cxml, expected_value = request.param
        stop = _GradientStop(element(gs_cxml))
        return stop, expected_value

    @pytest.fixture(
        params=[
            ("a:gs{pos=0}", 0.42, 0.42),
            ("a:gs{pos=42000}", 1.0, 1.0),
            ("a:gs{pos=100000}", 0.0, 0.0),
        ]
    )
    def pos_set_fixture(self, request):
        gs_cxml, new_value, expected_value = request.param
        stop = _GradientStop(element(gs_cxml))
        return stop, new_value, expected_value

    @pytest.fixture(params=[-0.42, 1.001])
    def raises_fixture(self, request):
        out_of_range_value = request.param
        stop = _GradientStop(element("a:gs{pos=50000}"))
        return stop, out_of_range_value



# encoding: utf-8

"""
Data label-related objects.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from ..text.text import Font, TextFrame
from ..util import lazyproperty


class DataLabels(object):
    """Provides access to properties of data labels for a plot or a series.

    This is not a collection and does not provide access to individual data
    labels. Access to individual labels is via the |Point| object. The
    properties this object provides control formatting of *all* the data
    labels in its scope.
    """

    def __init__(self, dLbls):
        super(DataLabels, self).__init__()
        self._element = dLbls

    @lazyproperty
    def font(self):
        """
        The |Font| object that provides access to the text properties for
        these data labels, such as bold, italic, etc.
        """
        defRPr = self._element.defRPr
        font = Font(defRPr)
        return font

    @property
    def number_format(self):
        """
        Read/write string specifying the format for the numbers on this set
        of data labels. Returns 'General' if no number format has been set.
        Note that this format string has no effect on rendered data labels
        when :meth:`number_format_is_linked` is |True|. Assigning a format
        string to this property automatically sets
        :meth:`number_format_is_linked` to |False|.
        """
        numFmt = self._element.numFmt
        if numFmt is None:
            return "General"
        return numFmt.formatCode

    @number_format.setter
    def number_format(self, value):
        self._element.get_or_add_numFmt().formatCode = value
        self.number_format_is_linked = False

    @property
    def number_format_is_linked(self):
        """
        Read/write boolean specifying whether number formatting should be
        taken from the source spreadsheet rather than the value of
        :meth:`number_format`.
        """
        numFmt = self._element.numFmt
        if numFmt is None:
            return True
        souceLinked = numFmt.sourceLinked
        if souceLinked is None:
            return True
        return numFmt.sourceLinked

    @number_format_is_linked.setter
    def number_format_is_linked(self, value):
        numFmt = self._element.get_or_add_numFmt()
        numFmt.sourceLinked = value

    @property
    def position(self):
        """
        Read/write :ref:`XlDataLabelPosition` enumeration value specifying
        the position of the data labels with respect to their data point, or
        |None| if no position is specified. Assigning |None| causes
        PowerPoint to choose the default position, which varies by chart
        type.
        """
        dLblPos = self._element.dLblPos
        if dLblPos is None:
            return None
        return dLblPos.val

    @position.setter
    def position(self, value):
        if value is None:
            self._element._remove_dLblPos()
            return
        self._element.get_or_add_dLblPos().val = value

    @property
    def show_category_name(self):
        """Read/write. True when name of category should appear in label."""
        return self._element.get_or_add_showCatName().val

    @show_category_name.setter
    def show_category_name(self, value):
        self._element.get_or_add_showCatName().val = bool(value)

    @property
    def show_legend_key(self):
        """Read/write. True when data label displays legend-color swatch."""
        return self._element.get_or_add_showLegendKey().val

    @show_legend_key.setter
    def show_legend_key(self, value):
        self._element.get_or_add_showLegendKey().val = bool(value)

    @property
    def show_percentage(self):
        """Read/write. True when data label displays percentage.

        This option is not operative on all chart types. Percentage appears
        on polar charts such as pie and donut.
        """
        return self._element.get_or_add_showPercent().val

    @show_percentage.setter
    def show_percentage(self, value):
        self._element.get_or_add_showPercent().val = bool(value)

    @property
    def show_series_name(self):
        """Read/write. True when data label displays series name."""
        return self._element.get_or_add_showSerName().val

    @show_series_name.setter
    def show_series_name(self, value):
        self._element.get_or_add_showSerName().val = bool(value)

    @property
    def show_value(self):
        """Read/write. True when label displays numeric value of datapoint."""
        return self._element.get_or_add_showVal().val

    @show_value.setter
    def show_value(self, value):
        self._element.get_or_add_showVal().val = bool(value)


class DataLabel(object):
    """
    The data label associated with an individual data point.
    """

    def __init__(self, ser, idx):
        super(DataLabel, self).__init__()
        self._ser = self._element = ser
        self._idx = idx

    @lazyproperty
    def font(self):
        """The |Font| object providing text formatting for this data label.

        This font object is used to customize the appearance of automatically
        inserted text, such as the data point value. The font applies to the
        entire data label. More granular control of the appearance of custom
        data label text is controlled by a font object on runs in the text
        frame.
        """
        txPr = self._get_or_add_txPr()
        text_frame = TextFrame(txPr, self)
        paragraph = text_frame.paragraphs[0]
        return paragraph.font

    @property
    def has_text_frame(self):
        """
        Return |True| if this data label has a text frame (implying it has
        custom data label text), and |False| otherwise. Assigning |True|
        causes a text frame to be added if not already present. Assigning
        |False| causes any existing text frame to be removed along with any
        text contained in the text frame.
        """
        dLbl = self._dLbl
        if dLbl is None:
            return False
        if dLbl.xpath("c:tx/c:rich"):
            return True
        return False

    @has_text_frame.setter
    def has_text_frame(self, value):
        if bool(value) is True:
            self._get_or_add_tx_rich()
        else:
            self._remove_tx_rich()

    @property
    def position(self):
        """
        Read/write :ref:`XlDataLabelPosition` member specifying the position
        of this data label with respect to its data point, or |None| if no
        position is specified. Assigning |None| causes PowerPoint to choose
        the default position, which varies by chart type.
        """
        dLbl = self._dLbl
        if dLbl is None:
            return None
        dLblPos = dLbl.dLblPos
        if dLblPos is None:
            return None
        return dLblPos.val

    @position.setter
    def position(self, value):
        if value is None:
            dLbl = self._dLbl
            if dLbl is None:
                return
            dLbl._remove_dLblPos()
            return
        dLbl = self._get_or_add_dLbl()
        dLbl.get_or_add_dLblPos().val = value

    @property
    def text_frame(self):
        """
        |TextFrame| instance for this data label, containing the text of the
        data label and providing access to its text formatting properties.
        """
        rich = self._get_or_add_rich()
        return TextFrame(rich, self)

    @property
    def _dLbl(self):
        """
        Return the |CT_DLbl| instance referring specifically to this
        individual data label (having the same index value), or |None| if not
        present.
        """
        return self._ser.get_dLbl(self._idx)

    def _get_or_add_dLbl(self):
        """
        The ``CT_DLbl`` instance referring specifically to this individual
        data label, newly created if not yet present in the XML.
        """
        return self._ser.get_or_add_dLbl(self._idx)

    def _get_or_add_rich(self):
        """
        Return the `c:rich` element representing the text frame for this data
        label, newly created with its ancestors if not present.
        """
        dLbl = self._get_or_add_dLbl()

        # having a c:spPr or c:txPr when a c:tx is present causes the "can't
        # save" bug on bubble charts. Remove c:spPr and c:txPr when present.
        dLbl._remove_spPr()
        dLbl._remove_txPr()

        return dLbl.get_or_add_rich()

    def _get_or_add_tx_rich(self):
        """
        Return the `c:tx` element for this data label, with its `c:rich`
        child and descendants, newly created if not yet present.
        """
        dLbl = self._get_or_add_dLbl()

        # having a c:spPr or c:txPr when a c:tx is present causes the "can't
        # save" bug on bubble charts. Remove c:spPr and c:txPr when present.
        dLbl._remove_spPr()
        dLbl._remove_txPr()

        return dLbl.get_or_add_tx_rich()

    def _get_or_add_txPr(self):
        """Return the `c:txPr` element for this data label.

        The `c:txPr` element and its parent `c:dLbl` element are created if
        not yet present.
        """
        dLbl = self._get_or_add_dLbl()
        return dLbl.get_or_add_txPr()

    def _remove_tx_rich(self):
        """
        Remove any `c:tx/c:rich` child of the `c:dLbl` element for this data
        label. Do nothing if that element is not present.
        """
        dLbl = self._dLbl
        if dLbl is None:
            return
        dLbl.remove_tx_rich()



# encoding: utf-8

"""
Test suite for pptx.presentation module.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from pptx.parts.coreprops import CorePropertiesPart
from pptx.parts.presentation import PresentationPart
from pptx.parts.slide import NotesMasterPart
from pptx.presentation import Presentation
from pptx.slide import SlideLayouts, SlideMaster, SlideMasters, Slides

from .unitutil.cxml import element, xml
from .unitutil.mock import class_mock, instance_mock, property_mock


class DescribePresentation(object):
    def it_knows_the_height_of_its_slides(self, sld_height_get_fixture):
        prs, expected_value = sld_height_get_fixture
        assert prs.slide_height == expected_value

    def it_can_change_the_height_of_its_slides(self, sld_height_set_fixture):
        prs, slide_height, expected_xml = sld_height_set_fixture
        prs.slide_height = slide_height
        assert prs._element.xml == expected_xml

    def it_knows_the_width_of_its_slides(self, sld_width_get_fixture):
        prs, expected_value = sld_width_get_fixture
        assert prs.slide_width == expected_value

    def it_can_change_the_width_of_its_slides(self, sld_width_set_fixture):
        prs, slide_width, expected_xml = sld_width_set_fixture
        prs.slide_width = slide_width
        assert prs._element.xml == expected_xml

    def it_knows_its_part(self, part_fixture):
        prs, prs_part_ = part_fixture
        assert prs.part is prs_part_

    def it_provides_access_to_its_core_properties(self, core_props_fixture):
        prs, core_properties_ = core_props_fixture
        assert prs.core_properties is core_properties_

    def it_provides_access_to_its_notes_master(self, notes_master_fixture):
        prs, notes_master_ = notes_master_fixture
        assert prs.notes_master is notes_master_

    def it_provides_access_to_its_slides(self, slides_fixture):
        prs, rename_slide_parts_, rIds = slides_fixture[:3]
        Slides_, slides_, expected_xml = slides_fixture[3:]
        slides = prs.slides
        rename_slide_parts_.assert_called_once_with(rIds)
        Slides_.assert_called_once_with(prs._element.xpath("p:sldIdLst")[0], prs)
        assert prs._element.xml == expected_xml
        assert slides is slides_

    def it_provides_access_to_its_slide_layouts(self, layouts_fixture):
        prs, slide_layouts_ = layouts_fixture
        assert prs.slide_layouts is slide_layouts_

    def it_provides_access_to_its_slide_master(self, master_fixture):
        prs, getitem_, slide_master_ = master_fixture
        slide_master = prs.slide_master
        getitem_.assert_called_once_with(0)
        assert slide_master is slide_master_

    def it_provides_access_to_its_slide_masters(self, masters_fixture):
        prs, SlideMasters_, slide_masters_, expected_xml = masters_fixture
        slide_masters = prs.slide_masters
        SlideMasters_.assert_called_once_with(
            prs._element.xpath("p:sldMasterIdLst")[0], prs
        )
        assert slide_masters is slide_masters_
        assert prs._element.xml == expected_xml

    def it_can_save_the_presentation_to_a_file(self, save_fixture):
        prs, file_, prs_part_ = save_fixture
        prs.save(file_)
        prs_part_.save.assert_called_once_with(file_)

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def core_props_fixture(self, prs_part_, core_properties_):
        prs = Presentation(None, prs_part_)
        prs_part_.core_properties = core_properties_
        return prs, core_properties_

    @pytest.fixture
    def layouts_fixture(self, masters_prop_, slide_layouts_):
        prs = Presentation(None, None)
        masters_prop_.return_value.__getitem__.return_value.slide_layouts = (
            slide_layouts_
        )
        return prs, slide_layouts_

    @pytest.fixture
    def master_fixture(self, masters_prop_, slide_master_):
        prs = Presentation(None, None)
        getitem_ = masters_prop_.return_value.__getitem__
        getitem_.return_value = slide_master_
        return prs, getitem_, slide_master_

    @pytest.fixture(
        params=[
            ("p:presentation", "p:presentation/p:sldMasterIdLst"),
            ("p:presentation/p:sldMasterIdLst", "p:presentation/p:sldMasterIdLst"),
        ]
    )
    def masters_fixture(self, request, SlideMasters_, slide_masters_):
        prs_cxml, expected_cxml = request.param
        prs = Presentation(element(prs_cxml), None)
        expected_xml = xml(expected_cxml)
        return prs, SlideMasters_, slide_masters_, expected_xml

    @pytest.fixture
    def notes_master_fixture(self, prs_part_, notes_master_):
        prs = Presentation(None, prs_part_)
        prs_part_.notes_master = notes_master_
        return prs, notes_master_

    @pytest.fixture
    def part_fixture(self, prs_part_):
        prs = Presentation(None, prs_part_)
        return prs, prs_part_

    @pytest.fixture
    def save_fixture(self, prs_part_):
        prs = Presentation(None, prs_part_)
        file_ = "foobar.docx"
        return prs, file_, prs_part_

    @pytest.fixture(
        params=[("p:presentation", None), ("p:presentation/p:sldSz{cy=42}", 42)]
    )
    def sld_height_get_fixture(self, request):
        prs_cxml, expected_value = request.param
        prs = Presentation(element(prs_cxml), None)
        return prs, expected_value

    @pytest.fixture(
        params=[
            ("p:presentation", "p:presentation/p:sldSz{cy=914400}"),
            ("p:presentation/p:sldSz{cy=424242}", "p:presentation/p:sldSz{cy=914400}"),
        ]
    )
    def sld_height_set_fixture(self, request):
        prs_cxml, expected_cxml = request.param
        prs = Presentation(element(prs_cxml), None)
        expected_xml = xml(expected_cxml)
        return prs, 914400, expected_xml

    @pytest.fixture(
        params=[("p:presentation", None), ("p:presentation/p:sldSz{cx=42}", 42)]
    )
    def sld_width_get_fixture(self, request):
        prs_cxml, expected_value = request.param
        prs = Presentation(element(prs_cxml), None)
        return prs, expected_value

    @pytest.fixture(
        params=[
            ("p:presentation", "p:presentation/p:sldSz{cx=914400}"),
            ("p:presentation/p:sldSz{cx=424242}", "p:presentation/p:sldSz{cx=914400}"),
        ]
    )
    def sld_width_set_fixture(self, request):
        prs_cxml, expected_cxml = request.param
        prs = Presentation(element(prs_cxml), None)
        expected_xml = xml(expected_cxml)
        return prs, 914400, expected_xml

    @pytest.fixture(
        params=[
            ("p:presentation", [], "p:presentation/p:sldIdLst"),
            (
                "p:presentation/p:sldIdLst/p:sldId{r:id=a}",
                ["a"],
                "p:presentation/p:sldIdLst/p:sldId{r:id=a}",
            ),
            (
                "p:presentation/p:sldIdLst/(p:sldId{r:id=a},p:sldId{r:id=b})",
                ["a", "b"],
                "p:presentation/p:sldIdLst/(p:sldId{r:id=a},p:sldId{r:id=b})",
            ),
        ]
    )
    def slides_fixture(self, request, part_prop_, Slides_, slides_):
        prs_cxml, rIds, expected_cxml = request.param
        prs = Presentation(element(prs_cxml), None)
        rename_slide_parts_ = part_prop_.return_value.rename_slide_parts
        expected_xml = xml(expected_cxml)
        return prs, rename_slide_parts_, rIds, Slides_, slides_, expected_xml

    # fixture components ---------------------------------------------

    @pytest.fixture
    def core_properties_(self, request):
        return instance_mock(request, CorePropertiesPart)

    @pytest.fixture
    def masters_prop_(self, request):
        return property_mock(request, Presentation, "slide_masters")

    @pytest.fixture
    def notes_master_(self, request):
        return instance_mock(request, NotesMasterPart)

    @pytest.fixture
    def part_prop_(self, request):
        return property_mock(request, Presentation, "part")

    @pytest.fixture
    def prs_part_(self, request):
        return instance_mock(request, PresentationPart)

    @pytest.fixture
    def slide_layouts_(self, request):
        return instance_mock(request, SlideLayouts)

    @pytest.fixture
    def SlideMasters_(self, request, slide_masters_):
        return class_mock(
            request, "pptx.presentation.SlideMasters", return_value=slide_masters_
        )

    @pytest.fixture
    def slide_master_(self, request):
        return instance_mock(request, SlideMaster)

    @pytest.fixture
    def slide_masters_(self, request):
        return instance_mock(request, SlideMasters)

    @pytest.fixture
    def Slides_(self, request, slides_):
        return class_mock(request, "pptx.presentation.Slides", return_value=slides_)

    @pytest.fixture
    def slides_(self, request):
        return instance_mock(request, Slides)



# encoding: utf-8

"""
Test suite for pptx.text.layout module
"""

from __future__ import absolute_import, print_function, unicode_literals

import pytest

from pptx.text.layout import _BinarySearchTree, _Line, _LineSource, TextFitter

from ..unitutil.mock import (
    call,
    class_mock,
    function_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class DescribeTextFitter(object):
    def it_can_determine_the_best_fit_font_size(self, best_fit_fixture):
        text, extents, max_size, font_file = best_fit_fixture[:4]
        _LineSource_, _init_, line_source_ = best_fit_fixture[4:7]
        _best_fit_font_size_, font_size_ = best_fit_fixture[7:]

        font_size = TextFitter.best_fit_font_size(text, extents, max_size, font_file)

        _LineSource_.assert_called_once_with(text)
        _init_.assert_called_once_with(line_source_, extents, font_file)
        _best_fit_font_size_.assert_called_once_with(max_size)
        assert font_size is font_size_

    def it_finds_best_fit_font_size_to_help_best_fit(self, _best_fit_fixture):
        text_fitter, max_size, _BinarySearchTree_ = _best_fit_fixture[:3]
        sizes_, predicate_, font_size_ = _best_fit_fixture[3:]

        font_size = text_fitter._best_fit_font_size(max_size)

        _BinarySearchTree_.from_ordered_sequence.assert_called_once_with(
            range(1, max_size + 1)
        )
        sizes_.find_max.assert_called_once_with(predicate_)
        assert font_size is font_size_

    def it_provides_a_fits_inside_predicate_fn(self, fits_pred_fixture):
        text_fitter, point_size = fits_pred_fixture[:2]
        _rendered_size_, expected_bool_value = fits_pred_fixture[2:]

        predicate = text_fitter._fits_inside_predicate
        result = predicate(point_size)

        text_fitter._wrap_lines.assert_called_once_with(
            text_fitter._line_source, point_size
        )
        _rendered_size_.assert_called_once_with(
            "Ty", point_size, text_fitter._font_file
        )
        assert result is expected_bool_value

    def it_provides_a_fits_in_width_predicate_fn(self, fits_cx_pred_fixture):
        text_fitter, point_size, line = fits_cx_pred_fixture[:3]
        _rendered_size_, expected_value = fits_cx_pred_fixture[3:]

        predicate = text_fitter._fits_in_width_predicate(point_size)
        result = predicate(line)

        _rendered_size_.assert_called_once_with(
            line.text, point_size, text_fitter._font_file
        )
        assert result is expected_value

    def it_wraps_lines_to_help_best_fit(self, wrap_fixture):
        text_fitter, line_source, point_size, remainder = wrap_fixture

        text_fitter._wrap_lines(line_source, point_size)

        assert text_fitter._break_line.call_args_list == [
            call(line_source, point_size),
            call(remainder, point_size),
        ]

    def it_breaks_off_a_line_to_help_wrap(self, break_fixture):
        text_fitter, line_source_, point_size = break_fixture[:3]
        _BinarySearchTree_, bst_, predicate_ = break_fixture[3:6]
        max_value_ = break_fixture[6]

        value = text_fitter._break_line(line_source_, point_size)

        _BinarySearchTree_.from_ordered_sequence.assert_called_once_with(line_source_)
        text_fitter._fits_in_width_predicate.assert_called_once_with(point_size)
        bst_.find_max.assert_called_once_with(predicate_)
        assert value is max_value_

    # fixtures ---------------------------------------------

    @pytest.fixture
    def best_fit_fixture(self, _LineSource_, _init_, _best_fit_font_size_):
        text, extents, max_size = "Foobar", (19, 20), 42
        font_file = "foobar.ttf"
        line_source_ = _LineSource_.return_value
        font_size_ = _best_fit_font_size_.return_value
        return (
            text,
            extents,
            max_size,
            font_file,
            _LineSource_,
            _init_,
            line_source_,
            _best_fit_font_size_,
            font_size_,
        )

    @pytest.fixture
    def _best_fit_fixture(self, _BinarySearchTree_, _fits_inside_predicate_):
        text_fitter = TextFitter(None, (None, None), None)
        max_size = 42
        sizes_ = _BinarySearchTree_.from_ordered_sequence.return_value
        predicate_ = _fits_inside_predicate_.return_value
        font_size_ = sizes_.find_max.return_value
        return (
            text_fitter,
            max_size,
            _BinarySearchTree_,
            sizes_,
            predicate_,
            font_size_,
        )

    @pytest.fixture
    def break_fixture(
        self, line_source_, _BinarySearchTree_, bst_, _fits_in_width_predicate_
    ):
        text_fitter = TextFitter(None, (None, None), None)
        point_size = 21
        _BinarySearchTree_.from_ordered_sequence.return_value = bst_
        predicate_ = _fits_in_width_predicate_.return_value
        max_value_ = bst_.find_max.return_value
        return (
            text_fitter,
            line_source_,
            point_size,
            _BinarySearchTree_,
            bst_,
            predicate_,
            max_value_,
        )

    @pytest.fixture(params=[(49, True), (50, True), (51, False)])
    def fits_cx_pred_fixture(self, request, _rendered_size_):
        rendered_width, expected_value = request.param
        text_fitter = TextFitter(None, (50, None), "foobar.ttf")
        point_size, line = 12, _Line("foobar", None)
        _rendered_size_.return_value = (rendered_width, None)
        return (text_fitter, point_size, line, _rendered_size_, expected_value)

    @pytest.fixture(
        params=[
            ((66, 99), 6, ("foo", "bar"), False),
            ((66, 100), 6, ("foo", "bar"), True),
            ((66, 101), 6, ("foo", "bar"), True),
        ]
    )
    def fits_pred_fixture(self, request, line_source_, _wrap_lines_, _rendered_size_):
        extents, point_size, text_lines, expected_value = request.param
        text_fitter = TextFitter(line_source_, extents, "foobar.ttf")
        _wrap_lines_.return_value = text_lines
        _rendered_size_.return_value = (None, 50)
        return text_fitter, point_size, _rendered_size_, expected_value

    @pytest.fixture
    def wrap_fixture(self, _break_line_):
        text_fitter = TextFitter(None, (None, None), None)
        point_size = 21
        line_source, remainder = _LineSource("foo bar"), _LineSource("bar")
        _break_line_.side_effect = [("foo", remainder), ("bar", _LineSource(""))]
        return text_fitter, line_source, point_size, remainder

    # fixture components -----------------------------------

    @pytest.fixture
    def _best_fit_font_size_(self, request):
        return method_mock(request, TextFitter, "_best_fit_font_size")

    @pytest.fixture
    def _BinarySearchTree_(self, request):
        return class_mock(request, "pptx.text.layout._BinarySearchTree")

    @pytest.fixture
    def _break_line_(self, request):
        return method_mock(request, TextFitter, "_break_line")

    @pytest.fixture
    def bst_(self, request):
        return instance_mock(request, _BinarySearchTree)

    @pytest.fixture
    def _fits_in_width_predicate_(self, request):
        return method_mock(request, TextFitter, "_fits_in_width_predicate")

    @pytest.fixture
    def _fits_inside_predicate_(self, request):
        return property_mock(request, TextFitter, "_fits_inside_predicate")

    @pytest.fixture
    def _init_(self, request):
        return initializer_mock(request, TextFitter)

    @pytest.fixture
    def _LineSource_(self, request, line_source_):
        return class_mock(
            request, "pptx.text.layout._LineSource", return_value=line_source_
        )

    @pytest.fixture
    def line_source_(self, request):
        return instance_mock(request, _LineSource)

    @pytest.fixture
    def _rendered_size_(self, request):
        return function_mock(request, "pptx.text.layout._rendered_size")

    @pytest.fixture
    def _wrap_lines_(self, request):
        return method_mock(request, TextFitter, "_wrap_lines")


class Describe_BinarySearchTree(object):
    def it_can_construct_from_an_ordered_sequence(self):
        bst = _BinarySearchTree.from_ordered_sequence(range(10))

        def in_order(node):
            """
            Traverse the tree depth first to produce a list of its values,
            in order.
            """
            result = []
            if node is None:
                return result
            result.extend(in_order(node._lesser))
            result.append(node.value)
            result.extend(in_order(node._greater))
            return result

        assert bst.value == 9
        assert bst._lesser.value == 4
        assert bst._greater is None
        assert in_order(bst) == list(range(10))

    def it_can_find_the_max_value_satisfying_a_predicate(self, max_fixture):
        bst, predicate, expected_value = max_fixture
        assert bst.find_max(predicate) == expected_value

    # fixtures ---------------------------------------------

    @pytest.fixture(
        params=[
            (range(10), lambda n: n < 6.5, 6),
            (range(10), lambda n: n > 9.9, None),
            (range(10), lambda n: n < 0.0, None),
        ]
    )
    def max_fixture(self, request):
        seq, predicate, expected_value = request.param
        bst = _BinarySearchTree.from_ordered_sequence(seq)
        return bst, predicate, expected_value


class Describe_LineSource(object):
    def it_generates_text_remainder_pairs(self):
        line_source = _LineSource("foo bar baz")
        expected = (
            ("foo", _LineSource("bar baz")),
            ("foo bar", _LineSource("baz")),
            ("foo bar baz", _LineSource("")),
        )
        assert all((a == b) for a, b in zip(expected, line_source))


# produces different results on Linux, fails Travis-CI

# from pptx.text.layout import _rendered_size
# from ..unitutil.file import testfile
# class Describe_rendered_size(object):

#     def it_calculates_the_rendered_size_of_text_at_point_size(self, fixture):
#         text, point_size, font_file, expected_value = fixture
#         extents = _rendered_size(text, point_size, font_file)
#         assert extents == expected_value

#     # fixtures ---------------------------------------------

#     @pytest.fixture(params=[
#         ('Typical',     18, (673100, 254000)),
#         ('foo bar baz', 12, (698500, 165100)),
#     ])
#     def fixture(self, request):
#         text, point_size, expected_value = request.param
#         font_file = testfile('calibriz.ttf')
#         return text, point_size, font_file, expected_value



# encoding: utf-8

"""Data point-related objects."""

from __future__ import absolute_import, division, print_function, unicode_literals

from pptx.chart.datalabel import DataLabel
from pptx.chart.marker import Marker
from pptx.compat import Sequence
from pptx.dml.chtfmt import ChartFormat
from pptx.util import lazyproperty


class _BasePoints(Sequence):
    """
    Sequence providing access to the individual data points in a series.
    """

    def __init__(self, ser):
        super(_BasePoints, self).__init__()
        self._element = ser
        self._ser = ser

    def __getitem__(self, idx):
        if idx < 0 or idx >= self.__len__():
            raise IndexError("point index out of range")
        return Point(self._ser, idx)


class BubblePoints(_BasePoints):
    """
    Sequence providing access to the individual data points in
    a |BubbleSeries| object.
    """

    def __len__(self):
        return min(
            self._ser.xVal_ptCount_val,
            self._ser.yVal_ptCount_val,
            self._ser.bubbleSize_ptCount_val,
        )


class CategoryPoints(_BasePoints):
    """
    Sequence providing access to individual |Point| objects, each
    representing the visual properties of a data point in the specified
    category series.
    """

    def __len__(self):
        return self._ser.cat_ptCount_val


class Point(object):
    """
    Provides access to the properties of an individual data point in
    a series, such as the visual properties of its marker and the text and
    font of its data label.
    """

    def __init__(self, ser, idx):
        super(Point, self).__init__()
        self._element = ser
        self._ser = ser
        self._idx = idx

    @lazyproperty
    def data_label(self):
        """
        The |DataLabel| object representing the label on this data point.
        """
        return DataLabel(self._ser, self._idx)

    @lazyproperty
    def format(self):
        """
        The |ChartFormat| object providing access to the shape formatting
        properties of this data point, such as line and fill.
        """
        dPt = self._ser.get_or_add_dPt_for_point(self._idx)
        return ChartFormat(dPt)

    @lazyproperty
    def marker(self):
        """
        The |Marker| instance for this point, providing access to the visual
        properties of the data point marker, such as fill and line. Setting
        these properties overrides any value set at the series level.
        """
        dPt = self._ser.get_or_add_dPt_for_point(self._idx)
        return Marker(dPt)


class XyPoints(_BasePoints):
    """
    Sequence providing access to the individual data points in an |XySeries|
    object.
    """

    def __len__(self):
        return min(self._ser.xVal_ptCount_val, self._ser.yVal_ptCount_val)



# encoding: utf-8

"""
Test suite for pptx.package module
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from pptx.media import Video
from pptx.opc.constants import RELATIONSHIP_TYPE as RT
from pptx.opc.package import Part, _Relationship
from pptx.opc.packuri import PackURI
from pptx.package import _ImageParts, _MediaParts, Package
from pptx.parts.coreprops import CorePropertiesPart
from pptx.parts.image import Image, ImagePart
from pptx.parts.media import MediaPart


from .unitutil.mock import call, class_mock, instance_mock, method_mock, property_mock


class DescribePackage(object):
    def it_provides_access_to_its_core_properties_part(self):
        pkg = Package.open("pptx/templates/default.pptx")
        assert isinstance(pkg.core_properties, CorePropertiesPart)

    def it_can_get_or_add_an_image_part(self, image_part_fixture):
        package, image_file, image_parts_, image_part_ = image_part_fixture
        image_part = package.get_or_add_image_part(image_file)
        image_parts_.get_or_add_image_part.assert_called_once_with(image_file)
        assert image_part is image_part_

    def it_can_get_or_add_a_media_part(self, media_part_fixture):
        package, media, media_part_ = media_part_fixture
        media_part = package.get_or_add_media_part(media)
        package._media_parts.get_or_add_media_part.assert_called_once_with(media)
        assert media_part is media_part_

    def it_knows_the_next_available_image_partname(self, next_fixture):
        package, ext, expected_value = next_fixture
        partname = package.next_image_partname(ext)
        assert partname == expected_value

    def it_knows_the_next_available_media_partname(self, nmp_fixture):
        package, ext, expected_value = nmp_fixture
        partname = package.next_media_partname(ext)
        assert partname == expected_value

    def it_provides_access_to_its_MediaParts_object(self, m_parts_fixture):
        package, _MediaParts_, media_parts_ = m_parts_fixture
        media_parts = package._media_parts
        _MediaParts_.assert_called_once_with(package)
        assert media_parts is media_parts_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def image_part_fixture(self, image_parts_, image_part_, _image_parts_prop_):
        package = Package()
        image_file = "foobar.png"
        _image_parts_prop_.return_value = image_parts_
        image_parts_.get_or_add_image_part.return_value = image_part_
        return package, image_file, image_parts_, image_part_

    @pytest.fixture
    def media_part_fixture(self, media_, media_part_, _media_parts_prop_, media_parts_):
        package = Package()
        _media_parts_prop_.return_value = media_parts_
        media_parts_.get_or_add_media_part.return_value = media_part_
        return package, media_, media_part_

    @pytest.fixture
    def m_parts_fixture(self, _MediaParts_, media_parts_):
        package = Package()
        _MediaParts_.return_value = media_parts_
        return package, _MediaParts_, media_parts_

    @pytest.fixture(params=[((3, 4, 2), 1), ((4, 2, 1), 3), ((2, 3, 1), 4)])
    def next_fixture(self, request, iter_parts_):
        idxs, idx = request.param
        package = Package()
        package.iter_parts.return_value = self.i_image_parts(request, idxs)
        ext = "foo"
        expected_value = "/ppt/media/image%d.%s" % (idx, ext)
        return package, ext, expected_value

    @pytest.fixture(params=[((3, 4, 2), 1), ((4, 2, 1), 3), ((2, 3, 1), 4)])
    def nmp_fixture(self, request, iter_parts_):
        idxs, idx = request.param
        package = Package()
        package.iter_parts.return_value = self.i_media_parts(request, idxs)
        ext = "foo"
        expected_value = "/ppt/media/media%d.%s" % (idx, ext)
        return package, ext, expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def image_part_(self, request):
        return instance_mock(request, ImagePart)

    @pytest.fixture
    def image_parts_(self, request):
        return instance_mock(request, _ImageParts)

    @pytest.fixture
    def _image_parts_prop_(self, request):
        return property_mock(request, Package, "_image_parts")

    def i_image_parts(self, request, idxs):
        def part(idx):
            partname = PackURI("/ppt/media/image%d.png" % idx)
            return instance_mock(request, Part, partname=partname)

        return iter([part(idx) for idx in idxs])

    def i_media_parts(self, request, idxs):
        def part(idx):
            partname = PackURI("/ppt/media/media%d.mp4" % idx)
            return instance_mock(request, Part, partname=partname)

        return iter([part(idx) for idx in idxs])

    @pytest.fixture
    def iter_parts_(self, request):
        return property_mock(request, Package, "iter_parts")

    @pytest.fixture
    def media_(self, request):
        return instance_mock(request, Video)

    @pytest.fixture
    def media_part_(self, request):
        return instance_mock(request, MediaPart)

    @pytest.fixture
    def _MediaParts_(self, request):
        return class_mock(request, "pptx.package._MediaParts")

    @pytest.fixture
    def media_parts_(self, request):
        return instance_mock(request, _MediaParts)

    @pytest.fixture
    def _media_parts_prop_(self, request):
        return property_mock(request, Package, "_media_parts")


class Describe_ImageParts(object):
    def it_can_iterate_over_the_package_image_parts(self, iter_fixture):
        image_parts, expected_parts = iter_fixture
        assert list(image_parts) == expected_parts

    def it_can_get_a_matching_image_part(self, get_fixture):
        image_parts, image_file, Image_, image_, image_part_ = get_fixture

        image_part = image_parts.get_or_add_image_part(image_file)

        Image_.from_file.assert_called_once_with(image_file)
        image_parts._find_by_sha1.assert_called_once_with(image_.sha1)
        assert image_part is image_part_

    def it_can_add_an_image_part(self, add_fixture):
        image_parts, image_file, Image_, image_ = add_fixture[:4]
        ImagePart_, package_, image_part_ = add_fixture[4:]

        image_part = image_parts.get_or_add_image_part(image_file)

        Image_.from_file.assert_called_once_with(image_file)
        image_parts._find_by_sha1.assert_called_once_with(image_.sha1)
        ImagePart_.new.assert_called_once_with(package_, image_)
        assert image_part is image_part_

    def it_can_find_an_image_part_by_sha1_hash(self, find_fixture):
        image_parts, sha1, expected_value = find_fixture
        image_part = image_parts._find_by_sha1(sha1)
        assert image_part is expected_value

    def but_it_skips_unsupported_image_types(self, request, _iter_):
        sha1 = "f00beed"
        svg_part_ = instance_mock(request, Part, name="svg_part_")
        png_part_ = instance_mock(request, ImagePart, name="png_part_", sha1=sha1)
        # ---order iteration to encounter svg part before target part---
        _iter_.return_value = iter((svg_part_, png_part_))
        image_parts = _ImageParts(None)

        result = image_parts._find_by_sha1(sha1)

        assert result == png_part_

    # fixtures ---------------------------------------------

    @pytest.fixture
    def add_fixture(
        self, package_, Image_, image_, _find_by_sha1_, ImagePart_, image_part_
    ):
        image_parts = _ImageParts(package_)
        image_file = "foobar.png"
        Image_.from_file.return_value = image_
        _find_by_sha1_.return_value = None
        ImagePart_.new.return_value = image_part_
        return (
            image_parts,
            image_file,
            Image_,
            image_,
            ImagePart_,
            package_,
            image_part_,
        )

    @pytest.fixture(params=[True, False])
    def find_fixture(self, request, _iter_, image_part_):
        image_part_is_present = request.param
        image_parts = _ImageParts(None)
        _iter_.return_value = iter((image_part_,))
        sha1 = "foobar"
        if image_part_is_present:
            image_part_.sha1 = "foobar"
            expected_value = image_part_
        else:
            image_part_.sha1 = "barfoo"
            expected_value = None
        return image_parts, sha1, expected_value

    @pytest.fixture
    def get_fixture(self, Image_, image_, image_part_, _find_by_sha1_):
        image_parts = _ImageParts(None)
        image_file = "foobar.png"
        Image_.from_file.return_value = image_
        _find_by_sha1_.return_value = image_part_
        return image_parts, image_file, Image_, image_, image_part_

    @pytest.fixture
    def iter_fixture(self, request, package_):
        def rel(is_external, reltype):
            part = instance_mock(request, Part)
            return instance_mock(
                request,
                _Relationship,
                is_external=is_external,
                reltype=reltype,
                target_part=part,
            )

        rels = (rel(True, RT.IMAGE), rel(False, RT.SLIDE), rel(False, RT.IMAGE))

        package_.iter_rels.return_value = iter((rels[0], rels[1], rels[2], rels[2]))
        image_parts = _ImageParts(package_)
        expected_parts = [rels[2].target_part]
        return image_parts, expected_parts

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _find_by_sha1_(self, request):
        return method_mock(request, _ImageParts, "_find_by_sha1")

    @pytest.fixture
    def Image_(self, request):
        return class_mock(request, "pptx.package.Image")

    @pytest.fixture
    def image_(self, request):
        return instance_mock(request, Image)

    @pytest.fixture
    def ImagePart_(self, request):
        return class_mock(request, "pptx.package.ImagePart")

    @pytest.fixture
    def image_part_(self, request):
        return instance_mock(request, ImagePart)

    @pytest.fixture
    def _iter_(self, request):
        return method_mock(request, _ImageParts, "__iter__")

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, Package)


class Describe_MediaParts(object):
    def it_can_iterate_the_media_parts_in_the_package(self, iter_fixture):
        media_parts, expected_parts = iter_fixture
        assert list(media_parts) == expected_parts

    def it_can_get_or_add_a_media_part(self, get_or_add_fixture):
        media_parts, media_, sha1, MediaPart_, calls = get_or_add_fixture[:5]
        media_part_ = get_or_add_fixture[5]

        media_part = media_parts.get_or_add_media_part(media_)

        media_parts._find_by_sha1.assert_called_once_with(media_parts, sha1)
        assert MediaPart_.new.call_args_list == calls
        assert media_part is media_part_

    def it_can_find_a_media_part_by_sha1(self, find_fixture):
        media_parts, sha1, expected_value = find_fixture
        media_part = media_parts._find_by_sha1(sha1)
        assert media_part is expected_value

    # fixtures ---------------------------------------------

    @pytest.fixture(params=[True, False])
    def find_fixture(self, request, _iter_, media_part_):
        media_part_is_present = request.param
        media_parts = _MediaParts(None)
        _iter_.return_value = iter((media_part_,))
        sha1 = "foobar"
        if media_part_is_present:
            media_part_.sha1 = "foobar"
            expected_value = media_part_
        else:
            media_part_.sha1 = "barfoo"
            expected_value = None
        return media_parts, sha1, expected_value

    @pytest.fixture(params=[True, False])
    def get_or_add_fixture(
        self, request, package_, media_, MediaPart_, media_part_, _find_by_sha1_
    ):
        media_present = request.param
        media_parts = _MediaParts(package_)
        media_.sha1 = sha1 = "2468"
        calls = [] if media_present else [call(package_, media_)]
        _find_by_sha1_.return_value = media_part_ if media_present else None
        MediaPart_.new.return_value = None if media_present else media_part_
        return media_parts, media_, sha1, MediaPart_, calls, media_part_

    @pytest.fixture
    def iter_fixture(self, request, package_):
        def rel(is_external, reltype, part):
            return instance_mock(
                request,
                _Relationship,
                is_external=is_external,
                reltype=reltype,
                target_part=part,
            )

        part_mocks = (
            instance_mock(request, Part, name="linked-media"),
            instance_mock(request, Part, name="slide"),
            instance_mock(request, Part, name="embeded-media"),
        )

        rels = (
            rel(True, RT.MEDIA, part_mocks[0]),
            rel(True, RT.VIDEO, part_mocks[0]),
            rel(False, RT.SLIDE, part_mocks[1]),
            rel(False, RT.MEDIA, part_mocks[2]),
            rel(False, RT.VIDEO, part_mocks[2]),
        )

        package_.iter_rels.return_value = iter(rels)

        media_parts = _MediaParts(package_)
        expected_parts = [part_mocks[2]]
        return media_parts, expected_parts

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _find_by_sha1_(self, request):
        return method_mock(request, _MediaParts, "_find_by_sha1", autospec=True)

    @pytest.fixture
    def _iter_(self, request):
        return method_mock(request, _MediaParts, "__iter__")

    @pytest.fixture
    def media_(self, request):
        return instance_mock(request, Video)

    @pytest.fixture
    def MediaPart_(self, request):
        return class_mock(request, "pptx.package.MediaPart")

    @pytest.fixture
    def media_part_(self, request):
        return instance_mock(request, MediaPart)

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, Package)



# encoding: utf-8

"""Slide-related custom element classes, including those for masters."""

from __future__ import absolute_import, division, print_function, unicode_literals

from pptx.oxml import parse_from_template, parse_xml
from pptx.oxml.dml.fill import CT_GradientFillProperties
from pptx.oxml.ns import nsdecls
from pptx.oxml.simpletypes import XsdString
from pptx.oxml.xmlchemy import (
    BaseOxmlElement,
    Choice,
    OneAndOnlyOne,
    OptionalAttribute,
    RequiredAttribute,
    ZeroOrMore,
    ZeroOrOne,
    ZeroOrOneChoice,
)


class _BaseSlideElement(BaseOxmlElement):
    """
    Base class for the six slide types, providing common methods.
    """

    @property
    def spTree(self):
        """
        Return required `p:cSld/p:spTree` grandchild.
        """
        return self.cSld.spTree


class CT_Background(BaseOxmlElement):
    """`p:bg` element."""

    # ---these two are actually a choice, not a sequence, but simpler for
    # ---present purposes this way.
    _tag_seq = ("p:bgPr", "p:bgRef")
    bgPr = ZeroOrOne("p:bgPr", successors=())
    bgRef = ZeroOrOne("p:bgRef", successors=())
    del _tag_seq

    def add_noFill_bgPr(self):
        """Return a new `p:bgPr` element with noFill properties."""
        xml = (
            "<p:bgPr %s>\n"
            "  <a:noFill/>\n"
            "  <a:effectLst/>\n"
            "</p:bgPr>" % nsdecls("a", "p")
        )
        bgPr = parse_xml(xml)
        self._insert_bgPr(bgPr)
        return bgPr


class CT_BackgroundProperties(BaseOxmlElement):
    """`p:bgPr` element."""

    _tag_seq = (
        "a:noFill",
        "a:solidFill",
        "a:gradFill",
        "a:blipFill",
        "a:pattFill",
        "a:grpFill",
        "a:effectLst",
        "a:effectDag",
        "a:extLst",
    )
    eg_fillProperties = ZeroOrOneChoice(
        (
            Choice("a:noFill"),
            Choice("a:solidFill"),
            Choice("a:gradFill"),
            Choice("a:blipFill"),
            Choice("a:pattFill"),
            Choice("a:grpFill"),
        ),
        successors=_tag_seq[6:],
    )
    del _tag_seq

    def _new_gradFill(self):
        """Override default to add default gradient subtree."""
        return CT_GradientFillProperties.new_gradFill()


class CT_CommonSlideData(BaseOxmlElement):
    """`p:cSld` element."""

    _tag_seq = ("p:bg", "p:spTree", "p:custDataLst", "p:controls", "p:extLst")
    bg = ZeroOrOne("p:bg", successors=_tag_seq[1:])
    spTree = OneAndOnlyOne("p:spTree")
    del _tag_seq
    name = OptionalAttribute("name", XsdString, default="")

    def get_or_add_bgPr(self):
        """Return `p:bg/p:bgPr` grandchild.

        If no such grandchild is present, any existing `p:bg` child is first
        removed and a new default `p:bg` with noFill settings is added.
        """
        bg = self.bg
        if bg is None or bg.bgPr is None:
            self._change_to_noFill_bg()
        return self.bg.bgPr

    def _change_to_noFill_bg(self):
        """Establish a `p:bg` child with no-fill settings.

        Any existing `p:bg` child is first removed.
        """
        self._remove_bg()
        bg = self.get_or_add_bg()
        bg.add_noFill_bgPr()
        return bg


class CT_NotesMaster(_BaseSlideElement):
    """
    ``<p:notesMaster>`` element, root of a notes master part
    """

    _tag_seq = ("p:cSld", "p:clrMap", "p:hf", "p:notesStyle", "p:extLst")
    cSld = OneAndOnlyOne("p:cSld")
    del _tag_seq

    @classmethod
    def new_default(cls):
        """
        Return a new ``<p:notesMaster>`` element based on the built-in
        default template.
        """
        return parse_from_template("notesMaster")


class CT_NotesSlide(_BaseSlideElement):
    """
    ``<p:notes>`` element, root of a notes slide part
    """

    _tag_seq = ("p:cSld", "p:clrMapOvr", "p:extLst")
    cSld = OneAndOnlyOne("p:cSld")
    del _tag_seq

    @classmethod
    def new(cls):
        """
        Return a new ``<p:notes>`` element based on the default template.
        Note that the template does not include placeholders, which must be
        subsequently cloned from the notes master.
        """
        return parse_from_template("notes")


class CT_Slide(_BaseSlideElement):
    """`p:sld` element, root element of a slide part (XML document)."""

    _tag_seq = ("p:cSld", "p:clrMapOvr", "p:transition", "p:timing", "p:extLst")
    cSld = OneAndOnlyOne("p:cSld")
    clrMapOvr = ZeroOrOne("p:clrMapOvr", successors=_tag_seq[2:])
    timing = ZeroOrOne("p:timing", successors=_tag_seq[4:])
    del _tag_seq

    @classmethod
    def new(cls):
        """Return new `p:sld` element configured as base slide shape."""
        return parse_xml(cls._sld_xml())

    @property
    def bg(self):
        """Return `p:bg` grandchild or None if not present."""
        return self.cSld.bg

    def get_or_add_childTnLst(self):
        """Return parent element for a new `p:video` child element.

        The `p:video` element causes play controls to appear under a video
        shape (pic shape containing video). There can be more than one video
        shape on a slide, which causes the precondition to vary. It needs to
        handle the case when there is no `p:sld/p:timing` element and when
        that element already exists. If the case isn't simple, it just nukes
        what's there and adds a fresh one. This could theoretically remove
        desired existing timing information, but there isn't any evidence
        available to me one way or the other, so I've taken the simple
        approach.
        """
        childTnLst = self._childTnLst
        if childTnLst is None:
            childTnLst = self._add_childTnLst()
        return childTnLst

    def _add_childTnLst(self):
        """Add `./p:timing/p:tnLst/p:par/p:cTn/p:childTnLst` descendant.

        Any existing `p:timing` child element is ruthlessly removed and
        replaced.
        """
        self.remove(self.get_or_add_timing())
        timing = parse_xml(self._childTnLst_timing_xml())
        self._insert_timing(timing)
        return timing.xpath("./p:tnLst/p:par/p:cTn/p:childTnLst")[0]

    @property
    def _childTnLst(self):
        """Return `./p:timing/p:tnLst/p:par/p:cTn/p:childTnLst` descendant.

        Return None if that element is not present.
        """
        childTnLsts = self.xpath("./p:timing/p:tnLst/p:par/p:cTn/p:childTnLst")
        if not childTnLsts:
            return None
        return childTnLsts[0]

    @staticmethod
    def _childTnLst_timing_xml():
        return (
            "<p:timing %s>\n"
            "  <p:tnLst>\n"
            "    <p:par>\n"
            '      <p:cTn id="1" dur="indefinite" restart="never" nodeType="'
            'tmRoot">\n'
            "        <p:childTnLst/>\n"
            "      </p:cTn>\n"
            "    </p:par>\n"
            "  </p:tnLst>\n"
            "</p:timing>" % nsdecls("p")
        )

    @staticmethod
    def _sld_xml():
        return (
            "<p:sld %s>\n"
            "  <p:cSld>\n"
            "    <p:spTree>\n"
            "      <p:nvGrpSpPr>\n"
            '        <p:cNvPr id="1" name=""/>\n'
            "        <p:cNvGrpSpPr/>\n"
            "        <p:nvPr/>\n"
            "      </p:nvGrpSpPr>\n"
            "      <p:grpSpPr/>\n"
            "    </p:spTree>\n"
            "  </p:cSld>\n"
            "  <p:clrMapOvr>\n"
            "    <a:masterClrMapping/>\n"
            "  </p:clrMapOvr>\n"
            "</p:sld>" % nsdecls("a", "p", "r")
        )


class CT_SlideLayout(_BaseSlideElement):
    """
    ``<p:sldLayout>`` element, root of a slide layout part
    """

    _tag_seq = ("p:cSld", "p:clrMapOvr", "p:transition", "p:timing", "p:hf", "p:extLst")
    cSld = OneAndOnlyOne("p:cSld")
    del _tag_seq


class CT_SlideLayoutIdList(BaseOxmlElement):
    """
    ``<p:sldLayoutIdLst>`` element, child of ``<p:sldMaster>`` containing
    references to the slide layouts that inherit from the slide master.
    """

    sldLayoutId = ZeroOrMore("p:sldLayoutId")


class CT_SlideLayoutIdListEntry(BaseOxmlElement):
    """
    ``<p:sldLayoutId>`` element, child of ``<p:sldLayoutIdLst>`` containing
    a reference to a slide layout.
    """

    rId = RequiredAttribute("r:id", XsdString)


class CT_SlideMaster(_BaseSlideElement):
    """
    ``<p:sldMaster>`` element, root of a slide master part
    """

    _tag_seq = (
        "p:cSld",
        "p:clrMap",
        "p:sldLayoutIdLst",
        "p:transition",
        "p:timing",
        "p:hf",
        "p:txStyles",
        "p:extLst",
    )
    cSld = OneAndOnlyOne("p:cSld")
    sldLayoutIdLst = ZeroOrOne("p:sldLayoutIdLst", successors=_tag_seq[3:])
    del _tag_seq


class CT_SlideTiming(BaseOxmlElement):
    """`p:timing` element, specifying animations and timed behaviors."""

    _tag_seq = ("p:tnLst", "p:bldLst", "p:extLst")
    tnLst = ZeroOrOne("p:tnLst", successors=_tag_seq[1:])
    del _tag_seq


class CT_TimeNodeList(BaseOxmlElement):
    """`p:tnLst` or `p:childTnList` element."""

    def add_video(self, shape_id):
        """Add a new `p:video` child element for movie having *shape_id*."""
        video_xml = (
            "<p:video %s>\n"
            '  <p:cMediaNode vol="80000">\n'
            '    <p:cTn id="%d" fill="hold" display="0">\n'
            "      <p:stCondLst>\n"
            '        <p:cond delay="indefinite"/>\n'
            "      </p:stCondLst>\n"
            "    </p:cTn>\n"
            "    <p:tgtEl>\n"
            '      <p:spTgt spid="%d"/>\n'
            "    </p:tgtEl>\n"
            "  </p:cMediaNode>\n"
            "</p:video>\n" % (nsdecls("p"), self._next_cTn_id, shape_id)
        )
        video = parse_xml(video_xml)
        self.append(video)

    @property
    def _next_cTn_id(self):
        """Return the next available unique ID (int) for p:cTn element."""
        cTn_id_strs = self.xpath("/p:sld/p:timing//p:cTn/@id")
        ids = [int(id_str) for id_str in cTn_id_strs]
        return max(ids) + 1


class CT_TLMediaNodeVideo(BaseOxmlElement):
    """`p:video` element, specifying video media details."""

    _tag_seq = ("p:cMediaNode",)
    cMediaNode = OneAndOnlyOne("p:cMediaNode")
    del _tag_seq



# encoding: utf-8

"""Test suite for pptx.part module."""

from __future__ import absolute_import

import pytest

from pptx.opc.constants import RELATIONSHIP_TYPE as RT
from pptx.opc.oxml import CT_Relationships
from pptx.opc.package import Part, _Relationship, RelationshipCollection
from pptx.opc.packuri import PackURI

from ..unitutil.mock import (
    call,
    class_mock,
    instance_mock,
    loose_mock,
    Mock,
    patch,
    PropertyMock,
)


class Describe_Relationship(object):
    def it_remembers_construction_values(self):
        # test data --------------------
        rId = "rId9"
        reltype = "reltype"
        target = Mock(name="target_part")
        external = False
        # exercise ---------------------
        rel = _Relationship(rId, reltype, target, None, external)
        # verify -----------------------
        assert rel.rId == rId
        assert rel.reltype == reltype
        assert rel.target_part == target
        assert rel.is_external == external

    def it_should_raise_on_target_part_access_on_external_rel(self):
        rel = _Relationship(None, None, None, None, external=True)
        with pytest.raises(ValueError):
            rel.target_part

    def it_should_have_target_ref_for_external_rel(self):
        rel = _Relationship(None, None, "target", None, external=True)
        assert rel.target_ref == "target"

    def it_should_have_relative_ref_for_internal_rel(self):
        """
        Internal relationships (TargetMode == 'Internal' in the XML) should
        have a relative ref, e.g. '../slideLayouts/slideLayout1.xml', for
        the target_ref attribute.
        """
        part = Mock(name="part", partname=PackURI("/ppt/media/image1.png"))
        baseURI = "/ppt/slides"
        rel = _Relationship(None, None, part, baseURI)  # external=False
        assert rel.target_ref == "../media/image1.png"


class DescribeRelationshipCollection(object):
    def it_also_has_dict_style_get_rel_by_rId(self, rels_with_known_rel):
        rels, rId, known_rel = rels_with_known_rel
        assert rels[rId] == known_rel

    def it_should_raise_on_failed_lookup_by_rId(self, rels):
        with pytest.raises(KeyError):
            rels["rId666"]

    def it_has_a_len(self, rels):
        assert len(rels) == 0

    def it_can_add_a_relationship(self, _Relationship_):
        baseURI, rId, reltype, target, is_external = (
            "baseURI",
            "rId9",
            "reltype",
            "target",
            False,
        )
        rels = RelationshipCollection(baseURI)
        rel = rels.add_relationship(reltype, target, rId, is_external)
        _Relationship_.assert_called_once_with(
            rId, reltype, target, baseURI, is_external
        )
        assert rels[rId] == rel
        assert rel == _Relationship_.return_value

    def it_can_add_a_relationship_if_not_found(
        self, rels_with_matching_rel_, rels_with_missing_rel_
    ):

        rels, reltype, part, matching_rel = rels_with_matching_rel_
        assert rels.get_or_add(reltype, part) == matching_rel

        rels, reltype, part, new_rel = rels_with_missing_rel_
        assert rels.get_or_add(reltype, part) == new_rel

    def it_knows_the_next_available_rId(self, rels_with_rId_gap):
        rels, expected_next_rId = rels_with_rId_gap
        next_rId = rels._next_rId
        assert next_rId == expected_next_rId

    def it_can_find_a_related_part_by_reltype(self, rels_with_target_known_by_reltype):
        rels, reltype, known_target_part = rels_with_target_known_by_reltype
        part = rels.part_with_reltype(reltype)
        assert part is known_target_part

    def it_can_find_a_related_part_by_rId(self, rels_with_known_target_part):
        rels, rId, known_target_part = rels_with_known_target_part
        part = rels.related_parts[rId]
        assert part is known_target_part

    def it_raises_KeyError_on_part_with_rId_not_found(self, rels):
        with pytest.raises(KeyError):
            rels.related_parts["rId666"]

    def it_can_compose_rels_xml(self, rels_with_known_rels, rels_elm):
        rels_with_known_rels.xml
        rels_elm.assert_has_calls(
            [
                call.add_rel("rId1", "http://rt-hyperlink", "http://some/link", True),
                call.add_rel("rId2", "http://rt-image", "../media/image1.png", False),
                call.xml(),
            ],
            any_order=True,
        )

    # def it_raises_on_add_rel_with_duplicate_rId(self, rels, rel):
    #     with pytest.raises(ValueError):
    #         rels.add_rel(rel)

    # fixtures ---------------------------------------------

    @pytest.fixture
    def _Relationship_(self, request):
        return class_mock(request, "pptx.opc.package._Relationship")

    @pytest.fixture
    def rel(self, _rId, _reltype, _target_part, _baseURI):
        return _Relationship(_rId, _reltype, _target_part, _baseURI)

    @pytest.fixture
    def rels(self, _baseURI):
        return RelationshipCollection(_baseURI)

    @pytest.fixture
    def rels_elm(self, request):
        """
        Return a rels_elm mock that will be returned from
        CT_Relationships.new()
        """
        # create rels_elm mock with a .xml property
        rels_elm = Mock(name="rels_elm")
        xml = PropertyMock(name="xml")
        type(rels_elm).xml = xml
        rels_elm.attach_mock(xml, "xml")
        rels_elm.reset_mock()  # to clear attach_mock call
        # patch CT_Relationships to return that rels_elm
        patch_ = patch.object(CT_Relationships, "new", return_value=rels_elm)
        patch_.start()
        request.addfinalizer(patch_.stop)
        return rels_elm

    @pytest.fixture
    def rels_with_known_rel(self, rels, _rId, rel):
        rels[_rId] = rel
        return rels, _rId, rel

    @pytest.fixture
    def rels_with_known_rels(self):
        """
        Populated RelationshipCollection instance that will exercise the
        rels.xml property.
        """
        rels = RelationshipCollection("/baseURI")
        rels.add_relationship(
            reltype="http://rt-hyperlink",
            target="http://some/link",
            rId="rId1",
            is_external=True,
        )
        part = Mock(name="part")
        part.partname.relative_ref.return_value = "../media/image1.png"
        rels.add_relationship(reltype="http://rt-image", target=part, rId="rId2")
        return rels

    @pytest.fixture
    def rels_with_known_target_part(self, rels, _rel_with_known_target_part):
        rel, rId, target_part = _rel_with_known_target_part
        rels.add_relationship(None, target_part, rId)
        return rels, rId, target_part

    @pytest.fixture
    def rels_with_matching_rel_(self, request, rels):
        matching_reltype_ = instance_mock(request, str, name="matching_reltype_")
        matching_part_ = instance_mock(request, Part, name="matching_part_")
        matching_rel_ = instance_mock(
            request,
            _Relationship,
            name="matching_rel_",
            reltype=matching_reltype_,
            target_part=matching_part_,
            is_external=False,
        )
        rels[1] = matching_rel_
        return rels, matching_reltype_, matching_part_, matching_rel_

    @pytest.fixture
    def rels_with_missing_rel_(self, request, rels, _Relationship_):
        missing_reltype_ = instance_mock(request, str, name="missing_reltype_")
        missing_part_ = instance_mock(request, Part, name="missing_part_")
        new_rel_ = instance_mock(
            request,
            _Relationship,
            name="new_rel_",
            reltype=missing_reltype_,
            target_part=missing_part_,
            is_external=False,
        )
        _Relationship_.return_value = new_rel_
        return rels, missing_reltype_, missing_part_, new_rel_

    @pytest.fixture
    def rels_with_rId_gap(self, request, rels):
        rel_with_rId1 = instance_mock(
            request, _Relationship, name="rel_with_rId1", rId="rId1"
        )
        rel_with_rId3 = instance_mock(
            request, _Relationship, name="rel_with_rId3", rId="rId3"
        )
        rels["rId1"] = rel_with_rId1
        rels["rId3"] = rel_with_rId3
        return rels, "rId2"

    @pytest.fixture
    def rels_with_target_known_by_reltype(
        self, rels, _rel_with_target_known_by_reltype
    ):
        rel, reltype, target_part = _rel_with_target_known_by_reltype
        rels[1] = rel
        return rels, reltype, target_part

    @pytest.fixture
    def _baseURI(self):
        return "/baseURI"

    @pytest.fixture
    def _rel_with_known_target_part(self, _rId, _reltype, _target_part, _baseURI):
        rel = _Relationship(_rId, _reltype, _target_part, _baseURI)
        return rel, _rId, _target_part

    @pytest.fixture
    def _rel_with_target_known_by_reltype(self, _rId, _reltype, _target_part, _baseURI):
        rel = _Relationship(_rId, _reltype, _target_part, _baseURI)
        return rel, _reltype, _target_part

    @pytest.fixture
    def _reltype(self):
        return RT.SLIDE

    @pytest.fixture
    def _rId(self):
        return "rId6"

    @pytest.fixture
    def _target_part(self, request):
        return loose_mock(request)



# encoding: utf-8

"""
lxml custom element classes for DrawingML-related XML elements.
"""

from __future__ import absolute_import

from pptx.enum.dml import MSO_PATTERN_TYPE
from pptx.oxml import parse_xml
from pptx.oxml.ns import nsdecls
from pptx.oxml.simpletypes import (
    ST_Percentage,
    ST_PositiveFixedAngle,
    ST_PositiveFixedPercentage,
    ST_RelationshipId,
)
from pptx.oxml.xmlchemy import (
    BaseOxmlElement,
    Choice,
    OneOrMore,
    OptionalAttribute,
    RequiredAttribute,
    ZeroOrOne,
    ZeroOrOneChoice,
)


class CT_Blip(BaseOxmlElement):
    """
    <a:blip> element
    """

    rEmbed = OptionalAttribute("r:embed", ST_RelationshipId)


class CT_BlipFillProperties(BaseOxmlElement):
    """
    Custom element class for <a:blipFill> element.
    """

    _tag_seq = ("a:blip", "a:srcRect", "a:tile", "a:stretch")
    blip = ZeroOrOne("a:blip", successors=_tag_seq[1:])
    srcRect = ZeroOrOne("a:srcRect", successors=_tag_seq[2:])
    del _tag_seq

    def crop(self, cropping):
        """
        Set `a:srcRect` child to crop according to *cropping* values.
        """
        srcRect = self._add_srcRect()
        srcRect.l, srcRect.t, srcRect.r, srcRect.b = cropping


class CT_GradientFillProperties(BaseOxmlElement):
    """`a:gradFill` custom element class."""

    _tag_seq = ("a:gsLst", "a:lin", "a:path", "a:tileRect")
    gsLst = ZeroOrOne("a:gsLst", successors=_tag_seq[1:])
    lin = ZeroOrOne("a:lin", successors=_tag_seq[2:])
    path = ZeroOrOne("a:path", successors=_tag_seq[3:])
    del _tag_seq

    @classmethod
    def new_gradFill(cls):
        """Return newly-created "loose" default gradient subtree."""
        return parse_xml(
            '<a:gradFill %s rotWithShape="1">\n'
            "  <a:gsLst>\n"
            '    <a:gs pos="0">\n'
            '      <a:schemeClr val="accent1">\n'
            '        <a:tint val="100000"/>\n'
            '        <a:shade val="100000"/>\n'
            '        <a:satMod val="130000"/>\n'
            "      </a:schemeClr>\n"
            "    </a:gs>\n"
            '    <a:gs pos="100000">\n'
            '      <a:schemeClr val="accent1">\n'
            '        <a:tint val="50000"/>\n'
            '        <a:shade val="100000"/>\n'
            '        <a:satMod val="350000"/>\n'
            "      </a:schemeClr>\n"
            "    </a:gs>\n"
            "  </a:gsLst>\n"
            '  <a:lin scaled="0"/>\n'
            "</a:gradFill>\n" % nsdecls("a")
        )

    def _new_gsLst(self):
        """Override default to add minimum subtree."""
        return CT_GradientStopList.new_gsLst()


class CT_GradientStop(BaseOxmlElement):
    """`a:gs` custom element class."""

    eg_colorChoice = ZeroOrOneChoice(
        (
            Choice("a:scrgbClr"),
            Choice("a:srgbClr"),
            Choice("a:hslClr"),
            Choice("a:sysClr"),
            Choice("a:schemeClr"),
            Choice("a:prstClr"),
        ),
        successors=(),
    )
    pos = RequiredAttribute("pos", ST_PositiveFixedPercentage)


class CT_GradientStopList(BaseOxmlElement):
    """`a:gsLst` custom element class."""

    gs = OneOrMore("a:gs")

    @classmethod
    def new_gsLst(cls):
        """Return newly-created "loose" default stop-list subtree.

        An `a:gsLst` element must have at least two `a:gs` children. These
        are the default from the PowerPoint built-in "White" template.
        """
        return parse_xml(
            "<a:gsLst %s>\n"
            '  <a:gs pos="0">\n'
            '    <a:schemeClr val="accent1">\n'
            '      <a:tint val="100000"/>\n'
            '      <a:shade val="100000"/>\n'
            '      <a:satMod val="130000"/>\n'
            "    </a:schemeClr>\n"
            "  </a:gs>\n"
            '  <a:gs pos="100000">\n'
            '    <a:schemeClr val="accent1">\n'
            '      <a:tint val="50000"/>\n'
            '      <a:shade val="100000"/>\n'
            '      <a:satMod val="350000"/>\n'
            "    </a:schemeClr>\n"
            "  </a:gs>\n"
            "</a:gsLst>\n" % nsdecls("a")
        )


class CT_GroupFillProperties(BaseOxmlElement):
    """`a:grpFill` custom element class"""


class CT_LinearShadeProperties(BaseOxmlElement):
    """`a:lin` custom element class"""

    ang = OptionalAttribute("ang", ST_PositiveFixedAngle)


class CT_NoFillProperties(BaseOxmlElement):
    """`a:noFill` custom element class"""


class CT_PatternFillProperties(BaseOxmlElement):
    """`a:pattFill` custom element class"""

    _tag_seq = ("a:fgClr", "a:bgClr")
    fgClr = ZeroOrOne("a:fgClr", successors=_tag_seq[1:])
    bgClr = ZeroOrOne("a:bgClr", successors=_tag_seq[2:])
    del _tag_seq
    prst = OptionalAttribute("prst", MSO_PATTERN_TYPE)

    def _new_bgClr(self):
        """Override default to add minimum subtree."""
        xml = (
            "<a:bgClr %s>\n" ' <a:srgbClr val="FFFFFF"/>\n' "</a:bgClr>\n"
        ) % nsdecls("a")
        bgClr = parse_xml(xml)
        return bgClr

    def _new_fgClr(self):
        """Override default to add minimum subtree."""
        xml = (
            "<a:fgClr %s>\n" ' <a:srgbClr val="000000"/>\n' "</a:fgClr>\n"
        ) % nsdecls("a")
        fgClr = parse_xml(xml)
        return fgClr


class CT_RelativeRect(BaseOxmlElement):
    """`a:srcRect` element and perhaps others."""

    l = OptionalAttribute("l", ST_Percentage, default=0.0)  # noqa
    t = OptionalAttribute("t", ST_Percentage, default=0.0)
    r = OptionalAttribute("r", ST_Percentage, default=0.0)
    b = OptionalAttribute("b", ST_Percentage, default=0.0)


class CT_SolidColorFillProperties(BaseOxmlElement):
    """`a:solidFill` custom element class."""

    eg_colorChoice = ZeroOrOneChoice(
        (
            Choice("a:scrgbClr"),
            Choice("a:srgbClr"),
            Choice("a:hslClr"),
            Choice("a:sysClr"),
            Choice("a:schemeClr"),
            Choice("a:prstClr"),
        ),
        successors=(),
    )
